"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const contro_1 = require("contro");
const engine_1 = require("./engine");
exports.Cell = engine_1.Cell;
exports.GameEngine = engine_1.GameEngine;
const game_1 = require("./models/game");
exports.GameData = game_1.GameData;
const parser_1 = __importDefault(require("./parser/parser"));
exports.Parser = parser_1.default;
const sounds_1 = require("./sounds");
exports.closeSounds = sounds_1.closeSounds;
exports.playSound = sounds_1.playSound;
const base_1 = __importDefault(require("./ui/base"));
exports.BaseUI = base_1.default;
const table_1 = __importDefault(require("./ui/table"));
exports.TableUI = table_1.default;
const util_1 = require("./util");
exports.RULE_DIRECTION = util_1.RULE_DIRECTION;
class TableEngine {
    constructor(table, customHandler) {
        const keyboard = new contro_1.Keyboard();
        this.gamepad = new contro_1.Gamepad();
        this.controls = {
            up: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.ARROW_UP), keyboard.key('w')) },
            down: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.ARROW_DOWN), keyboard.key('s')) },
            left: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.ARROW_LEFT), keyboard.key('a')) },
            right: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.ARROW_RIGHT), keyboard.key('d')) },
            action: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.CLUSTER_BOTTOM), keyboard.key('x'), keyboard.key('Space')) },
            undo: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.CLUSTER_LEFT), this.gamepad.button(contro_1.BUTTON_TYPE.BUMPER_TOP_LEFT), keyboard.key('z'), keyboard.key('u')) },
            restart: { lastPressed: null, button: contro_1.or(this.gamepad.button(contro_1.BUTTON_TYPE.CLUSTER_TOP), keyboard.key('r')) }
        };
        const makeChecker = (control, gameAction) => {
            return () => {
                const now = Date.now();
                const interval = this.tableUI.getGameData().metadata.keyRepeatInterval || .25;
                const { button, lastPressed } = this.controls[control];
                if (button.query()) {
                    if (!lastPressed || (lastPressed + interval * 1000) < now) {
                        gameAction();
                        this.controls[control].lastPressed = now;
                    }
                }
                else {
                    this.controls[control].lastPressed = null;
                }
            };
        };
        this.controlCheckers = [
            makeChecker('up', () => this.tableUI.pressUp()),
            makeChecker('down', () => this.tableUI.pressDown()),
            makeChecker('left', () => this.tableUI.pressLeft()),
            makeChecker('right', () => this.tableUI.pressRight()),
            makeChecker('action', () => this.tableUI.pressAction()),
            makeChecker('undo', () => this.tableUI.pressUndo()),
            makeChecker('restart', () => this.tableUI.pressRestart())
        ];
        this.tableUI = new table_1.default(table);
        this.timer = 0;
        this.currentLevel = 0;
        const defaultEventHandler = {
            onSound: (sound) => {
                // let sounds play while the game loads or player keeps moving
                sounds_1.playSound(sound); // tslint:disable-line:no-floating-promises
                return;
            },
            onLevelComplete: () => {
                if (!this.tableUI.isCurrentLevelAMessage()) {
                    alert(`Congratulations! You completed the level.`);
                }
            },
            onMessage: (message) => alert(message),
            onWin: () => alert(`You WON!`)
        };
        this.eventHandler = {
            onSound: (customHandler ? customHandler.onSound : null) || defaultEventHandler.onSound,
            onLevelComplete: (customHandler ? customHandler.onLevelComplete : null) || defaultEventHandler.onLevelComplete,
            onMessage: (customHandler ? customHandler.onMessage : null) || defaultEventHandler.onMessage,
            onWin: (customHandler ? customHandler.onWin : null) || defaultEventHandler.onWin
        };
    }
    setGame(source, levelNum) {
        this.tableUI.setGame(source);
        this.tableUI.setLevel(levelNum);
        this.currentLevel = levelNum;
    }
    start() {
        this.startTickHandler();
    }
    stop() {
        cancelAnimationFrame(this.timer);
    }
    pollControls() {
        this.controlCheckers.forEach((fn) => fn());
    }
    startTickHandler() {
        const runLoop = () => __awaiter(this, void 0, void 0, function* () {
            this.pollControls();
            this.tick();
            this.timer = window.requestAnimationFrame(runLoop);
        });
        this.timer = window.requestAnimationFrame(runLoop);
    }
    tick() {
        return __awaiter(this, void 0, void 0, function* () {
            while (this.tableUI.isCurrentLevelAMessage()) {
                yield this.eventHandler.onMessage(this.tableUI.getCurrentLevelMessage());
                this.currentLevel++;
                this.tableUI.setLevel(this.currentLevel);
            }
            const { 
            // changedCells,
            didLevelChange, didWinGame, messageToShow, soundToPlay
            // wasAgainTick
             } = this.tableUI.tick();
            if (soundToPlay) {
                yield this.eventHandler.onSound(soundToPlay);
            }
            if (didWinGame) {
                yield this.eventHandler.onWin();
                cancelAnimationFrame(this.timer);
                return; // make sure we don't call window.requestAnimationFrame again
            }
            else if (didLevelChange) {
                this.currentLevel += 1;
                this.tableUI.setLevel(this.currentLevel);
                yield this.eventHandler.onLevelComplete(this.currentLevel);
            }
            else if (messageToShow) {
                yield this.eventHandler.onMessage(messageToShow);
                this.tableUI.pressAction(); // Tell the engine we are ready to continue
            }
        });
    }
}
exports.TableEngine = TableEngine;
//# sourceMappingURL=index-browser.js.map